/* 
ResourceManagement.hpp

A template based simple and fast resource manager using STL and a base resource class to be used with the manager.

Copyright (c) 2008 by Ashic Mahtab and Zinat Wali. 
Permission is granted to use, distribute, or modify this source, provided that this copyright notice remains intact. 

The code is provided "as is" with no guarantees. 
It has been tested in Microsoft Visual Studio 2008 running under Windows Vista Ultimate x64. 

Feel free to contact us at:
aheartattack@gmail.com (Ashic)
zinat.wali@gmail.com (Zinat)

Happy coding :)
*/

#ifndef RESOURCEMANAGEMENT_HPP
#define RESOURCEMANAGEMENT_HPP
#include <string>
#include <vector>
#include <stack>

#define SAFE_DELETE( p )       { if( p ) { delete ( p );     ( p ) = NULL; } }


//-------------------------------------------------------------------------
// The Resource class.
//-------------------------------------------------------------------------
class Resource
{
protected:
	std::string	m_name;
	std::string	m_path;
	std::string	m_filename;
	unsigned int m_handle;
	unsigned long m_refCount;

public:
	//-------------------------------------------------------------------------
	// The Resource class constructor.
	//-------------------------------------------------------------------------
	Resource( const unsigned int handle, const std::string& name, const std::string& path = "./" )
	{
		// Store the name.
		if( !name.empty() )
			m_name = name;

		// Store the path.
		if( !path.empty() )
			m_path = path;

		//Store the handle
		m_handle = handle;

		// Create the filename.
		if( !name.empty() && !path.empty() )
			m_filename = path + name;

		// Start the reference count.
		m_refCount = 1;
	}

	//-------------------------------------------------------------------------
	// The Resource class destructor.
	//-------------------------------------------------------------------------
	virtual ~Resource()
	{
	}

	//-------------------------------------------------------------------------
	// Returns the name of the resource.
	//-------------------------------------------------------------------------
	std::string GetName()
	{
		return m_name;
	}

	//-------------------------------------------------------------------------
	// Returns the path to the resource.
	//-------------------------------------------------------------------------
	std::string GetPath()
	{
		return m_path;
	}

	//-------------------------------------------------------------------------
	// Returns the filename of the resource.
	//-------------------------------------------------------------------------
	std::string GetFilename()
	{
		return m_filename;
	}

	//-------------------------------------------------------------------------
	// Returns the handle of the resource.
	//-------------------------------------------------------------------------
	unsigned int GetHandle()
	{
		return m_handle;
	}

	//-------------------------------------------------------------------------
	// Increments the resource's reference count.
	//-------------------------------------------------------------------------
	void IncRef()
	{
		m_refCount++;
	}

	//-------------------------------------------------------------------------
	// Decrements the resource's reference count.
	//-------------------------------------------------------------------------
	void DecRef()
	{
		m_refCount--;
	}

	//-------------------------------------------------------------------------
	// Returns the resource's reference count.
	//-------------------------------------------------------------------------
	unsigned long GetRefCount()
	{
		return m_refCount;
	}

};

//-----------------------------------------------------------------------------
// Resource Manager Class
//-----------------------------------------------------------------------------
template <class T>
class ResourceManager
{
private:
	std::stack<unsigned int> m_handles;
	std::vector<T*> *m_list;	//List of pointers to resources
	void (*CreateResource)( T **resource, const unsigned int handle, const std::string& name, const std::string& path ); // Application specific resource creation.

public:
	//-------------------------------------------------------------------------
	// The resource manager class constructor.
	//-------------------------------------------------------------------------
	ResourceManager( void (*CreateResourceFunction)( T **resource, 
		const unsigned int handle, const std::string& name, const std::string& path ) = NULL )
	{
		m_list = new std::vector<T*>;

		CreateResource = CreateResourceFunction;
	}

	//-------------------------------------------------------------------------
	// The resource manager class destructor.
	//-------------------------------------------------------------------------
	~ResourceManager()
	{
		EmptyList();
		delete m_list;
	}

	//-------------------------------------------------------------------------
	// Returns a pointer to the vector of resources.
	//-------------------------------------------------------------------------
	std::vector<T*> *GetList()
	{
		return m_list;
	}

	//-------------------------------------------------------------------------
	// Returns a resource by its filename.
	//-------------------------------------------------------------------------
	T* GetElement( const std::string& name, const std::string& path = "./" )
	{
		// Ensure the name and path are valid, and the list is valid and not empty.
		if( name.empty() || path.empty() || m_list == NULL || m_list->empty())
			return NULL;

		//// Iterate the list looking for the specified resource.
		for(std::vector<T*>::iterator i = m_list->begin();
			i != m_list->end(); i++)
		{
			if((*i) != NULL)
				if((*i)->GetName() == name)
					if((*i)->GetPath() == path)
						return (*i);
		}

		// Return NULL if the resource was not found.
		return NULL;
	}

	//-------------------------------------------------------------------------
	// Returns a resource by its handle.
	//-------------------------------------------------------------------------
	T *GetElement( const int handle )
	{
		if(handle < m_list->size() && handle >= 0)
			return (*m_list)[handle];

		return NULL;
	}

	//-------------------------------------------------------------------------
	// Clears the resource list.
	//-------------------------------------------------------------------------
	void EmptyList()
	{
		for(std::vector<T*>::iterator i = m_list->begin();
			i != m_list->end(); i++)
			SAFE_DELETE(*i);

		while(!m_handles.empty())
			m_handles.pop();

		//use vector swapping trick to release previously used memory.
		m_list->clear();
		m_list->swap(*m_list);
	}

	//-------------------------------------------------------------------------
	// Removes the given resource from the manager.
	//-------------------------------------------------------------------------
	void Remove( const unsigned int handle )
	{
		// Ensure the resource to be removed and the list is valid.
		if(handle < 0 || m_list == NULL || handle >= m_list->size() || (*m_list)[handle] == NULL)
			return;

		T * resource = (*m_list)[handle];

		// Decrement the resource's reference count.
		resource->DecRef();

		// If the resource is no longer being used then destroy it.
		if( resource->GetRefCount() == 0 )
		{
			//add the handle to the stack of free handles
			m_handles.push(handle);
			//delete the resource
			delete resource;
			//mark the slot as NULL
			(*m_list)[handle] = NULL;
		}
	}

	//-------------------------------------------------------------------------
	// Adds a new resource to the manager.
	//-------------------------------------------------------------------------
	unsigned int Add( const std::string& name, const std::string& path = "./" )
	{
		// Ensure the list, the resource name, and its path are valid.
		if( m_list == NULL || name.empty() || path.empty() )
			return -1;

		// If the element already exists, then return a pointer to it.
		T *element = GetElement( name, path );
		if( element != NULL )
		{
			element->IncRef();
			return element->GetHandle();
		}


		//Check if there is an available handle. If not, use new handle.
		bool handleAvailable = !m_handles.empty();
		unsigned int handle;
		if(handleAvailable)
		{
			handle = m_handles.top();
			m_handles.pop();
		}
		else handle = m_list->size();

		// Create the resource, preferably through the application specific
		// function if it is available.
		T *resource = NULL;
		if( CreateResource != NULL )
			CreateResource( &resource, handle, name, path );
		else
			resource = new T( handle, name, path );


		//Add the resource to the manager. If there is an available handle, then 
		//we store the resource using the handle. Otherwise we add it to the vector.
		if(handleAvailable)
			(*m_list)[handle] = resource;
		else 
			m_list->push_back(resource);

		//return the handle
		return handle;
	}

	//-------------------------------------------------------------------------
	// Overloaded [] operator for easier access
	//-------------------------------------------------------------------------
	T* operator [](unsigned int handle)
	{
		if(handle < m_list->size() && handle >= 0)
			return (*m_list)[handle];

		return NULL;
	}
};

#endif